/*
 * Generic object event notification support.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <condition_variable>
#include <list>
#include <map>
#include <mutex>
#include <vector>

#include <resource/resource.h>
#include <systypes/base.h>

#include "notifier_interface.h"



/* Object-specific notification details. */

class ObjectNotificationState
{
public:
    /* Synchronisation primitives for state access and notification. */

    std::mutex lock;
    std::condition_variable condition;

    /* Pending notifications for monitored objects. */

    notify_flags_t pending_notifications = 0;
    notify_values_t pending_values = NOTIFY_VALUES_NULL;

    /* Endpoints associated with monitored objects. */

    l4_cap_idx_t endpoint = L4_INVALID_CAP;

    bool is_null() { return l4_is_invalid_cap(endpoint); }
};



/* Collection types. */

typedef std::map<notifiable_t *, ObjectNotificationState> ObjectNotificationStates;
typedef std::map<notifiable_t *, std::mutex> ObjectStateLocks;

typedef std::vector<ipc_server_config_type *> ServerConfigs;



/* An object for monitoring object event notifications. */

class ObjectNotifier
{
protected:
    /* General state access locking. */

    std::mutex _state_lock;

    /* Object-specific state locking. */

    ObjectStateLocks _object_locks;

    /* Notification state. */

    ObjectNotificationStates _state;

    /* Notifier resource details. */

    ServerConfigs _configs;
    bool _started = false;
    l4_cap_idx_t _irq = L4_INVALID_CAP;

    /* Convenience method to access object state. */

    virtual ObjectNotificationState &object_state(notifiable_t *object, bool create);

    /* Helper methods. */

    virtual bool _transfer(ObjectNotificationState &state, notifiable_t *object);

public:
    virtual ~ObjectNotifier();

    /* Local operations. */

    virtual long start();

    virtual void stop();

    virtual long subscribe(notifiable_t *object, notify_flags_t flags);

    virtual long unsubscribe(notifiable_t *object);

    virtual long remove_endpoint(notifiable_t *object, l4_cap_idx_t endpoint);

    /* Helper methods. */

    virtual void notify(notifiable_t *object, notify_flags_t flags,
                        notify_values_t values) = 0;
};



/* An object monitoring notifications for a collection of different objects. */

class GeneralObjectNotifier : public ObjectNotifier
{
protected:
    /* Locking to protect pending notification members and to coordinate access
       to notifications. */

    std::mutex _general_lock;

    /* General lock synchronisation. */

    std::condition_variable _general_condition;

    /* Objects affected by notifications. */

    std::list<notifiable_t *> _affected;

    /* Helper methods. */

    virtual bool _retrieve(notifiable_t **object);

    virtual bool _retrieve_for_object(notifiable_t *object);

public:
    virtual long wait(notifiable_t **object);

    /* Helper methods. */

    virtual void notify(notifiable_t *object, notify_flags_t flags,
                        notify_values_t values);
};



/* An object monitoring notifications for specific objects. */

class SpecificObjectNotifier : public ObjectNotifier
{
public:
    virtual long wait_object(notifiable_t *object);

    /* Helper methods. */

    virtual void notify(notifiable_t *object, notify_flags_t flags,
                        notify_values_t values);
};



/* Object-specific notifier resource. */

class NotifierResource : public Notifier, public Resource
{
protected:
    ObjectNotifier *_notifier;
    notifiable_t *_object;

public:
    explicit NotifierResource(ObjectNotifier *notifier, notifiable_t *object)
    : _notifier(notifier), _object(object)
    {
    }

    explicit NotifierResource()
    : _notifier(NULL), _object(NULL)
    {
    }

    /* Server details. */

    virtual ipc_server_default_config_type config();

    virtual void *interface()
    { return static_cast<Notifier *>(this); }

    /* Notifier methods. */

    virtual long notify(notify_flags_t flags, notify_values_t values);
};



/* Helper functions. */

SpecificObjectNotifier *notifier_get_task_notifier();

GeneralObjectNotifier *notifier_get_local_notifier();

// vim: tabstop=4 expandtab shiftwidth=4
